@@ -164,6 +164,20 @@ class CLIParser:
                 help="the CSV file(s) from which to read task data",
                 metavar="<data.csv>",
             )
+            parser.add_argument(
+                "--process-csv-only",
+                "-p",
+                action="store_true",
+                required=False,
+                help="process CSV variable substitutions only and output intermediate JSON",
+            )
+            parser.add_argument(
+                "--yes",
+                "-y",
+                action="store_true",
+                required=False,
+                help="proceed without requiring user confirmation",
+            )
 
         if any(
             module in sys.argv[0]
@@ -339,6 +353,15 @@ class CLIParser:
                 metavar="<compute_requirement.json>",
             )
 
+        if "admin" in sys.argv[0]:
+            parser.add_argument(
+                "work_requirement_id",
+                metavar="<work_requirement_id>",
+                type=str,
+                nargs="*",
+                help="work requirement to be refreshed",
+            )
+
         self.args = parser.parse_args()
 
         if self.args.docs:
@@ -499,6 +522,14 @@ class CLIParser:
     def csv_files(self) -> Optional[List[str]]:
         return self.args.csv_file
 
+    @property
+    def process_csv_only(self) -> Optional[bool]:
+        return self.args.process_csv_only
+
+    @property
+    def wr_ids(self) -> Optional[List[str]]:
+        return self.args.work_requirement_id
+
 
 def lookup_module_description(module_name: str) -> Optional[str]:
     """
