@@ -10,6 +10,8 @@ CAPTURE_TASKOUTPUT = "captureTaskOutput"  # Bool
 COMMON_SECTION = "common"
 COMPLETED_TASK_TTL = "completedTaskTtl"  # Float
 COMPUTE_REQUIREMENT_BATCH_SIZE = "computeRequirementBatchSize"  # Integer
+CSV_FILE = "csvFile"  # String
+CSV_FILES = "csvFiles"  # List of Strings
 DEPENDENT_ON = "dependentOn"  # String
 DOCKER_ENV = "dockerEnvironment"  # Dictionary
 DOCKER_PASSWORD = "dockerPassword"  # String
@@ -22,11 +24,13 @@ FINISH_IF_ANY_TASK_FAILED = "finishIfAnyTaskFailed"  # Boolean
 FLATTEN_PATHS = "flattenInputPaths"  # Boolean
 FLATTEN_UPLOAD_PATHS = "flattenUploadPaths"  # Boolean
 FULFIL_ON_SUBMIT = "fulfilOnSubmit"  # Boolean
+IMAGES_ID = "imagesId"  # String
 IMPORT = "import"  # String
-INITIAL_NODES = "targetInstanceCount"  # Integer
 INPUT_FILES = "inputs"  # List of Strings
+INSTANCE_TAGS = "instanceTags"  # List of Strings
 INSTANCE_TYPES = "instanceTypes"  # List of Strings
 KEY = "key"  # String
+MAINTAIN_INSTANCE_COUNT = "maintainInstanceCount"  # Bool
 MAX_NODES = "maxNodes"  # Integer
 MAX_RETRIES = "maximumTaskRetries"  # Integer
 MAX_WORKERS = "maxWorkers"  # Integer
@@ -43,14 +47,21 @@ PROVIDERS = "providers"  # List of Strings
 RAM = "ram"  # List of two Floats
 REGIONS = "regions"  # List of Strings
 SECRET = "secret"  # String
+TARGET_INSTANCE_COUNT = "targetInstanceCount"  # Integer
 TASKS = "tasks"  # List of Tasks
 TASKS_PER_WORKER = "tasksPerWorker"  # Integer
 TASK_COUNT = "taskCount"  # Integer
 TASK_GROUPS = "taskGroups"  # List of Task Groups
+TASK_GROUP_NAME = "taskGroupName"  # String
+TASK_NAME = "taskName"  # String
 TASK_TYPE = "taskType"  # String
 TASK_TYPES = "taskTypes"  # List of Strings
 TEMPLATE_ID = "templateId"  # String
+UPLOAD_FILES = "uploadFiles"  # List of Dicts
+LOCAL_PATH = "localPath"  # String
+UPLOAD_PATH = "uploadPath"  # String
 URL = "url"  # String
+USERDATA = "userData"  # String
 VCPUS = "vcpus"  # List of two Floats
 VERIFY_AT_START = "verifyAtStart"  # List of Strings
 VERIFY_WAIT = "verifyWait"  # List of Strings
@@ -78,6 +89,8 @@ ALL_KEYS = [
     COMMON_SECTION,
     COMPLETED_TASK_TTL,
     COMPUTE_REQUIREMENT_BATCH_SIZE,
+    CSV_FILE,
+    CSV_FILES,
     DEPENDENT_ON,
     DOCKER_ENV,
     DOCKER_PASSWORD,
@@ -90,11 +103,13 @@ ALL_KEYS = [
     FLATTEN_PATHS,
     FLATTEN_UPLOAD_PATHS,
     FULFIL_ON_SUBMIT,
+    IMAGES_ID,
     IMPORT,
-    INITIAL_NODES,
     INPUT_FILES,
+    INSTANCE_TAGS,
     INSTANCE_TYPES,
     KEY,
+    MAINTAIN_INSTANCE_COUNT,
     MAX_NODES,
     MAX_RETRIES,
     MAX_WORKERS,
@@ -110,14 +125,21 @@ ALL_KEYS = [
     RAM,
     REGIONS,
     SECRET,
+    TARGET_INSTANCE_COUNT,
     TASKS,
     TASKS_PER_WORKER,
     TASK_COUNT,
     TASK_GROUPS,
+    TASK_GROUP_NAME,
+    TASK_NAME,
     TASK_TYPE,
     TASK_TYPES,
     TEMPLATE_ID,
+    UPLOAD_FILES,
+    LOCAL_PATH,
+    UPLOAD_PATH,
     URL,
+    USERDATA,
     VCPUS,
     VERIFY_AT_START,
     VERIFY_WAIT,
