@@ -5,12 +5,21 @@ Module for handling Task data supplied in a CSV file
 import csv
 import re
 from ast import literal_eval
+from collections import OrderedDict
 from json import load as json_load
 from typing import Dict, List, Optional
 
+from toml import load as toml_load
+
 from yd_commands.args import ARGS_PARSER
+from yd_commands.config import ConfigWorkRequirement
 from yd_commands.config_keys import *
-from yd_commands.mustache import BOOL_SUB, NUMBER_SUB, process_mustache_substitutions
+from yd_commands.mustache import (
+    BOOL_SUB,
+    NUMBER_SUB,
+    load_jsonnet_file_with_mustache_substitutions,
+    process_mustache_substitutions,
+)
 from yd_commands.printing import print_json, print_log
 
 
@@ -74,6 +83,41 @@ class CSVTaskData:
         return self._total_tasks - self._index
 
 
+class CSVDataCache:
+    """
+    Caches CSV data to prevent multiple loads of the same CSV file
+    """
+
+    def __init__(self, max_entries: Optional[int] = None):
+        """
+        'max_entries' limits the size of the cache.
+        - Use default 'None' for unlimited caching
+        - Set to zero to disable caching
+        """
+        self._max_entries: Optional[int] = max_entries
+        self._csv_task_data_objects: OrderedDict[str, CSVTaskData] = OrderedDict()
+
+    def get_csv_task_data(self, csv_filename: str) -> CSVTaskData:
+        csv_task_data = self._csv_task_data_objects.get(csv_filename, None)
+        if csv_task_data:  # Cache hit
+            csv_task_data.reset()
+        else:  # Cache miss
+            if self._max_entries is not None:
+                if (
+                    len(self._csv_task_data_objects) == self._max_entries
+                    and len(self._csv_task_data_objects) > 0
+                ):
+                    self._csv_task_data_objects.popitem(last=False)
+            csv_task_data = CSVTaskData(csv_filename)
+            if self._max_entries != 0:
+                self._csv_task_data_objects[csv_filename] = csv_task_data
+        return csv_task_data
+
+
+# Singleton instance of the CSVDataCache class
+CSV_DATA_CACHE = CSVDataCache(max_entries=2)
+
+
 def load_json_file_with_csv_task_expansion(
     json_file: str, csv_files: List[str]
 ) -> Dict:
@@ -88,6 +132,33 @@ def load_json_file_with_csv_task_expansion(
     return perform_csv_task_expansion(wr_data, csv_files)
 
 
+def load_jsonnet_file_with_csv_task_expansion(
+    jsonnet_file: str, csv_files: List[str]
+) -> Dict:
+    """
+    Load a Jsonnet file, expanding its Task lists using data from CSV
+    files. Return the expanded and Mustache-processed Work Requirement data.
+    """
+
+    wr_data = load_jsonnet_file_with_mustache_substitutions(jsonnet_file)
+    return perform_csv_task_expansion(wr_data, csv_files)
+
+
+def load_toml_file_with_csv_task_expansion(
+    toml_file: str, csv_files: List[str]
+) -> Dict:
+    """
+    Load a TOML file Work Requirement, expanding its Task lists using data
+    from CSV files. Return the expanded and Mustache-processed Work Requirement
+    data.
+    """
+
+    with open(toml_file, "r") as f:
+        wr_data = toml_load(f)
+
+    return perform_csv_task_expansion(wr_data, csv_files)
+
+
 def perform_csv_task_expansion(wr_data: Dict, csv_files: List[str]) -> Dict:
     """
     Expand a Work Requirement using CSV data.
@@ -103,7 +174,7 @@ def perform_csv_task_expansion(wr_data: Dict, csv_files: List[str]) -> Dict:
         raise Exception("Number of CSV files exceeds number of Task Groups")
 
     for counter, csv_file in enumerate(csv_files):
-        csv_file, index = _get_csv_file_index(csv_file, wr_data[TASK_GROUPS])
+        csv_file, index = get_csv_file_index(csv_file, wr_data[TASK_GROUPS])
         if index is None:
             index = counter
 
@@ -117,10 +188,10 @@ def perform_csv_task_expansion(wr_data: Dict, csv_files: List[str]) -> Dict:
                 "when using CSV file for data"
             )
 
-        csv_data = CSVTaskData(csv_file)
+        csv_data = CSV_DATA_CACHE.get_csv_task_data(csv_file)
         task_prototype = task_group[TASKS][0]
 
-        if not _substitions_present(csv_data.var_names, str(task_prototype)):
+        if not substitions_present(csv_data.var_names, str(task_prototype)):
             print_log(
                 f"Warning: No CSV substitutions to apply to Task Group "
                 f"{index + 1}; not expanding Task list"
@@ -130,15 +201,13 @@ def perform_csv_task_expansion(wr_data: Dict, csv_files: List[str]) -> Dict:
         generated_task_list = []
         for task_data in csv_data:
             generated_task_list.append(
-                _csv_mustache_substitution(
-                    task_prototype, csv_data.var_names, task_data
-                )
+                csv_mustache_substitution(task_prototype, csv_data.var_names, task_data)
             )
         task_group[TASKS] = generated_task_list
         print_log(f"Generated {len(generated_task_list)} Task(s) from CSV data")
 
     if ARGS_PARSER.process_csv_only:
-        print_log("Processing CSV substitutions only")
+        print_log("Displaying CSV substitutions only:")
         print_json(wr_data)
         exit(0)
 
@@ -147,7 +216,7 @@ def perform_csv_task_expansion(wr_data: Dict, csv_files: List[str]) -> Dict:
     return wr_data
 
 
-def _csv_mustache_substitution(
+def csv_mustache_substitution(
     task_prototype: Dict, csv_var_names: List, task_data: List
 ) -> Dict:
     """
@@ -157,11 +226,11 @@ def _csv_mustache_substitution(
     subs_dict = {var_name: task_data[i] for i, var_name in enumerate(csv_var_names)}
     new_task = str(task_prototype)
     for var_name, value in subs_dict.items():
-        new_task = _make_string_substitutions(new_task, var_name, value)
+        new_task = make_string_substitutions(new_task, var_name, value)
     return literal_eval(new_task)  # Convert back from string
 
 
-def _make_string_substitutions(input: str, var_name: str, value: str) -> str:
+def make_string_substitutions(input: str, var_name: str, value: str) -> str:
     """
     Helper function to make string substitutions for CSV variables only.
     """
@@ -191,7 +260,7 @@ def _make_string_substitutions(input: str, var_name: str, value: str) -> str:
 USED_FILE_INDEXES = []
 
 
-def _get_csv_file_index(
+def get_csv_file_index(
     csv_filename: str, task_groups: List[Dict]
 ) -> [str, Optional[int]]:
     """
@@ -236,12 +305,50 @@ def _get_csv_file_index(
     return csv_filename, None
 
 
-def _substitions_present(var_names: List[str], task_prototype: str) -> bool:
+def substitions_present(var_names: List[str], task_prototype: str) -> bool:
     """
     Check if there are any CSV substitutions present in the Task prototype.
     """
     return (
         any(f"{{{{{var_name}}}}}" in task_prototype for var_name in var_names)
-        or any(f"{{{{num:{var_name}}}}}" in task_prototype for var_name in var_names)
-        or any(f"{{{{bool:{var_name}}}}}" in task_prototype for var_name in var_names)
+        or any(
+            f"{{{{{NUMBER_SUB}{var_name}}}}}" in task_prototype
+            for var_name in var_names
+        )
+        or any(
+            f"{{{{{BOOL_SUB}{var_name}}}}}" in task_prototype for var_name in var_names
+        )
     )
+
+
+def csv_expand_toml_tasks(config_wr: ConfigWorkRequirement, csv_file: str) -> Dict:
+    """
+    When there's a CSV file specified, but no JSON file, create the expanded
+    list of Tasks using the CSV data.
+    """
+    wr_data = {TASK_GROUPS: [{TASKS: [{}]}]}
+    task_proto = wr_data[TASK_GROUPS][0][TASKS][0]
+    csv_data = CSV_DATA_CACHE.get_csv_task_data(csv_file.split(":")[0])
+    for config_value, config_name in [
+        (config_wr.args, ARGS),
+        (config_wr.bash_script, BASH_SCRIPT),
+        (config_wr.capture_taskoutput, CAPTURE_TASKOUTPUT),
+        (config_wr.docker_env, DOCKER_ENV),
+        (config_wr.docker_password, DOCKER_PASSWORD),
+        (config_wr.docker_username, DOCKER_USERNAME),
+        (config_wr.env, ENV),
+        (config_wr.executable, EXECUTABLE),
+        (config_wr.flatten_input_paths, FLATTEN_PATHS),
+        (config_wr.input_files, INPUT_FILES),
+        (config_wr.output_files, OUTPUT_FILES),
+        (config_wr.task_name, TASK_NAME),
+        (config_wr.task_type, TASK_TYPE),
+        (config_wr.verify_at_start, VERIFY_AT_START),
+        (config_wr.verify_wait, VERIFY_WAIT),
+    ]:
+        if config_value is not None and substitions_present(
+            csv_data.var_names, str(config_value)
+        ):
+            task_proto[config_name] = config_value
+
+    return perform_csv_task_expansion(wr_data, [csv_file])
