@@ -19,8 +19,6 @@ from yellowdog_client.model import (
     CloudProvider,
     DoubleRange,
     FlattenPath,
-    ObjectPath,
-    ObjectPathsRequest,
     RunSpecification,
     Task,
     TaskGroup,
@@ -42,8 +40,10 @@ from yd_commands.config import (
 )
 from yd_commands.config_keys import *
 from yd_commands.csv_data import (
+    csv_expand_toml_tasks,
     load_json_file_with_csv_task_expansion,
-    perform_csv_task_expansion,
+    load_jsonnet_file_with_csv_task_expansion,
+    load_toml_file_with_csv_task_expansion,
 )
 from yd_commands.interactive import confirmed
 from yd_commands.mustache import (
@@ -63,6 +63,7 @@ from yd_commands.printing import (
     print_log,
     print_numbered_strings,
 )
+from yd_commands.submit_utils import UploadedFiles, generate_task_input_list
 from yd_commands.type_check import (
     check_bool,
     check_dict,
@@ -71,11 +72,11 @@ from yd_commands.type_check import (
     check_list,
     check_str,
 )
-from yd_commands.upload_utils import unique_upload_pathname, upload_file
+from yd_commands.upload_utils import unique_upload_pathname
 from yd_commands.validate_properties import validate_properties
 from yd_commands.wrapper import CLIENT, CONFIG_COMMON, main_wrapper
 
-# Import the configuration from the TOML file
+# Import the Work Requirement configuration from the TOML file
 CONFIG_WR: ConfigWorkRequirement = load_config_work_requirement()
 
 
@@ -86,6 +87,8 @@ INPUT_FOLDER_NAME = None
 if ARGS_PARSER.dry_run:
     WR_SNAPSHOT = WorkRequirementSnapshot()
 
+UPLOADED_FILES: Optional[UploadedFiles] = None
+
 
 @main_wrapper
 def main():
@@ -103,8 +106,13 @@ def main():
         CONFIG_WR.csv_files if ARGS_PARSER.csv_files is None else ARGS_PARSER.csv_files
     )
 
+    if not csv_files and ARGS_PARSER.process_csv_only:
+        raise Exception(
+            "Option '--process-csv-only' is only valid if CSV file(s) specified"
+        )
+
     if wr_data_file is None and csv_files is not None:
-        tasks_data = csv_expand_toml_tasks(csv_files[0])
+        tasks_data = csv_expand_toml_tasks(CONFIG_WR, csv_files[0])
         submit_work_requirement(
             directory_to_upload_from=CONFIG_FILE_DIR,
             tasks_data=tasks_data,
@@ -112,6 +120,8 @@ def main():
 
     elif wr_data_file is not None:
         print_log(f"Loading Work Requirement data from: '{wr_data_file}'")
+
+        # JSON file
         if wr_data_file.lower().endswith("json"):
             if csv_files is not None:
                 tasks_data = load_json_file_with_csv_task_expansion(
@@ -120,14 +130,32 @@ def main():
                 )
             else:
                 tasks_data = load_json_file_with_mustache_substitutions(wr_data_file)
-        elif wr_data_file.lower().endswith("toml"):
-            tasks_data = load_toml_file_with_mustache_substitutions(wr_data_file)
+
+        # Jsonnet file
         elif wr_data_file.lower().endswith("jsonnet"):
-            tasks_data = load_jsonnet_file_with_mustache_substitutions(wr_data_file)
+            if csv_files is not None:
+                tasks_data = load_jsonnet_file_with_csv_task_expansion(
+                    jsonnet_file=wr_data_file,
+                    csv_files=csv_files,
+                )
+            else:
+                tasks_data = load_jsonnet_file_with_mustache_substitutions(wr_data_file)
+
+        # TOML file (undocumented)
+        elif wr_data_file.lower().endswith("toml"):
+            if csv_files is not None:
+                tasks_data = load_toml_file_with_csv_task_expansion(
+                    toml_file=wr_data_file,
+                    csv_files=csv_files,
+                )
+            else:
+                tasks_data = load_toml_file_with_mustache_substitutions(wr_data_file)
+
+        # None of the above
         else:
             raise Exception(
                 f"Work Requirement data file '{wr_data_file}' "
-                "must end with '.json', '.toml' or 'jsonnet'"
+                "must end with '.json', '.jsonnet', or '.toml'"
             )
         validate_properties(tasks_data, "Work Requirement JSON")
         submit_work_requirement(
@@ -176,6 +204,10 @@ def submit_work_requirement(
     global ID
     ID = tasks_data.get(WR_NAME, ID if CONFIG_WR.wr_name is None else CONFIG_WR.wr_name)
 
+    # Handle any files that need to be uploaded
+    global UPLOADED_FILES
+    UPLOADED_FILES = UploadedFiles(client=CLIENT, wr_name=ID, config=CONFIG_COMMON)
+
     # Ensure we're in the correct directory for uploads
     if directory_to_upload_from != "":
         chdir(directory_to_upload_from)
@@ -187,13 +219,8 @@ def submit_work_requirement(
 
     # Create the list of Task Groups
     task_groups: List[TaskGroup] = []
-    input_files_by_task_group: List[List[str]] = []
     for tg_number, task_group_data in enumerate(tasks_data[TASK_GROUPS]):
-        task_group, input_files = create_task_group(
-            tg_number, tasks_data, task_group_data
-        )
-        task_groups.append(task_group)
-        input_files_by_task_group.append(input_files)
+        task_groups.append(create_task_group(tg_number, tasks_data, task_group_data))
 
     # Create the Work Requirement
     priority = check_float_or_int(tasks_data.get(PRIORITY, CONFIG_WR.priority))
@@ -219,37 +246,15 @@ def submit_work_requirement(
         global WR_SNAPSHOT
         WR_SNAPSHOT.set_work_requirement(work_requirement)
 
-    # Keep track of uploaded files
-    uploaded_files = []
-
     # Add Tasks to their Task Groups
     for tg_number, task_group in enumerate(task_groups):
         try:
-            if not ARGS_PARSER.dry_run:
-                # Upload files required by the Tasks in this Task Group
-                if len(input_files_by_task_group[tg_number]) > 0:
-                    print_log(f"Uploading files for Task Group '{task_group.name}'")
-                for input_file in input_files_by_task_group[tg_number]:
-                    if input_file not in uploaded_files:
-                        upload_file(
-                            client=CLIENT,
-                            filename=input_file,
-                            namespace=CONFIG_COMMON.namespace,
-                            id=ID,
-                            url=CONFIG_COMMON.url,
-                            input_folder_name=INPUT_FOLDER_NAME,
-                            flatten_upload_paths=flatten_upload_paths,
-                        )
-                        uploaded_files.append(input_file)
-
-            # Add the Tasks
             add_tasks_to_task_group(
                 tg_number,
                 task_group,
                 tasks_data,
                 task_count,
                 work_requirement,
-                uploaded_files,
                 flatten_upload_paths=flatten_upload_paths,
             )
 
@@ -268,10 +273,9 @@ def create_task_group(
     tg_number: int,
     tasks_data: Dict,
     task_group_data: Dict,
-) -> (TaskGroup, List[str]):
+) -> TaskGroup:
     """
-    Create a Task Group and return the list of unique input files required by
-    the Tasks in the Task Group.
+    Create a TaskGroup object.
     """
 
     # Remap 'task_type' to 'task_types' in the Task Group if 'task_types'
@@ -281,38 +285,26 @@ def create_task_group(
             task_group_data[TASK_TYPES] = [task_group_data[TASK_TYPE]]
     task_types_from_tasks = set()
 
-    # Gather input files and task types
-    input_files = []
+    # Gather task types
     for task in task_group_data[TASKS]:
-        input_files += check_list(
-            task.get(
-                INPUT_FILES,
-                task_group_data.get(
-                    INPUT_FILES, tasks_data.get(INPUT_FILES, CONFIG_WR.input_files)
-                ),
-            )
-        )
         try:
             task_types_from_tasks.add(task[TASK_TYPE])
         except KeyError:
             pass
 
-    # Deduplicate
-    input_files = sorted(list(set(input_files)))
-
     # Name the Task Group
     num_task_groups = len(tasks_data[TASK_GROUPS])
     num_tasks = len(task_group_data[TASKS])
     task_group_name = get_task_group_name(
-        task_group_data.get(NAME, None),
+        task_group_data.get(NAME, CONFIG_WR.task_group_name),
         tg_number,
         num_task_groups,
         num_tasks,
     )
 
-    # Assemble the RunSpecification values for the Task Group
-    # task_types can be automatically added to by the task_types
-    # specified in the Tasks
+    # Assemble the RunSpecification values for the Task Group;
+    # 'task_types' can automatically be added to by the task_types
+    # specified in the Tasks.
     task_types: List = list(
         set(
             check_list(task_group_data.get(TASK_TYPES, tasks_data.get(TASK_TYPES, [])))
@@ -393,7 +385,7 @@ def create_task_group(
     )
     completed_task_ttl = None if ctttl_data is None else timedelta(minutes=ctttl_data)
 
-    # Create the fully-populated Task Group
+    # Create the Task Group
     task_group = TaskGroup(
         name=task_group_name,
         runSpecification=run_specification,
@@ -419,8 +411,9 @@ def create_task_group(
         ),
         completedTaskTtl=completed_task_ttl,
     )
+
     print_log(f"Generated Task Group '{task_group_name}'")
-    return task_group, input_files
+    return task_group
 
 
 def add_tasks_to_task_group(
@@ -429,11 +422,10 @@ def add_tasks_to_task_group(
     tasks_data: Dict,
     task_count: Optional[int],
     work_requirement: WorkRequirement,
-    uploaded_files: List[str],
     flatten_upload_paths: bool = False,
 ) -> None:
     """
-    Adds all the Tasks that comprise a given Task Group
+    Add all the constituent Tasks to the Task Group.
     """
 
     # Ensure there's at least one Task
@@ -442,7 +434,7 @@ def add_tasks_to_task_group(
         tasks_data[TASK_GROUPS][tg_number][TASKS] = [{}]
         num_tasks = 1
 
-    # If 'taskCount' is set at the Task Group level, and there are
+    # If 'taskCount' is set at the Task Group level, and there
     # is only one Task, create 'taskCount' copies of the Task
     task_group_task_count = tasks_data[TASK_GROUPS][tg_number].get(TASK_COUNT, None)
     if task_group_task_count is not None:
@@ -469,7 +461,8 @@ def add_tasks_to_task_group(
 
     # Iterate through batches
     for batch_number in range(num_task_batches):
-        tasks_list.clear()
+
+        # Iterate through tasks in the batch
         for task_number in range(
             (TASK_BATCH_SIZE * batch_number),
             min(TASK_BATCH_SIZE * (batch_number + 1), num_tasks),
@@ -477,7 +470,11 @@ def add_tasks_to_task_group(
             task_group_data = tasks_data[TASK_GROUPS][tg_number]
             task = tasks[task_number] if task_count is None else tasks[0]
             task_name = get_task_name(
-                task.get(NAME, None), task_number, num_tasks, tg_number, num_task_groups
+                task.get(NAME, CONFIG_WR.task_name),
+                task_number,
+                num_tasks,
+                tg_number,
+                num_task_groups,
             )
             executable = check_str(
                 task.get(
@@ -531,35 +528,52 @@ def add_tasks_to_task_group(
                 input_files_list, verify_at_start_files_list, verify_wait_files_list
             )
 
-            input_files = [
-                TaskInput.from_task_namespace(
-                    unique_upload_pathname(
-                        filename=file,
-                        id=ID,
-                        input_folder_name=INPUT_FOLDER_NAME,
-                        flatten_upload_paths=flatten_upload_paths,
-                    ),
-                    verification=TaskInputVerification.VERIFY_AT_START,
+            # Upload files in the 'inputs' list
+            # (Duplicates won't be re-added)
+            for file in input_files_list:
+                UPLOADED_FILES.add_input_file(
+                    filename=file, flatten_upload_paths=flatten_upload_paths
                 )
-                for file in input_files_list
-            ]
-            verify_at_start = [
-                TaskInput.from_task_namespace(
-                    f"{ID}/{file}", verification=TaskInputVerification.VERIFY_AT_START
-                )
-                for file in verify_at_start_files_list
-            ]
-            input_files += verify_at_start
-            verify_wait = [
-                TaskInput.from_task_namespace(
-                    f"{ID}/{file}", verification=TaskInputVerification.VERIFY_WAIT
-                )
-                for file in verify_wait_files_list
-            ]
-            input_files += verify_wait
 
-            # Set up output files
-            output_files = [
+            # Upload files in the 'uploadFiles' list
+            upload_files = check_list(
+                task.get(
+                    UPLOAD_FILES,
+                    task_group_data.get(
+                        UPLOAD_FILES,
+                        tasks_data.get(UPLOAD_FILES, CONFIG_WR.upload_files),
+                    ),
+                )
+            )
+            for file in upload_files:
+                try:
+                    UPLOADED_FILES.add_upload_file(file[LOCAL_PATH], file[UPLOAD_PATH])
+                except KeyError:
+                    raise Exception(
+                        f"Property '{UPLOAD_FILES}' must have entries with "
+                        f"properties '{LOCAL_PATH}' and '{UPLOAD_PATH}'"
+                    )
+
+            # Set up the 'inputs' property
+            inputs = generate_task_input_list(
+                files=[
+                    unique_upload_pathname(
+                        input_file, ID, INPUT_FOLDER_NAME, False, flatten_upload_paths
+                    )
+                    for input_file in input_files_list
+                ],
+                verification=TaskInputVerification.VERIFY_AT_START,
+                wr_name=None,
+            )
+            inputs += generate_task_input_list(
+                verify_at_start_files_list, TaskInputVerification.VERIFY_AT_START, ID
+            )
+            inputs += generate_task_input_list(
+                verify_wait_files_list, TaskInputVerification.VERIFY_WAIT, ID
+            )
+
+            # Set up the 'outputs' property
+            outputs = [
                 TaskOutput.from_worker_directory(file)
                 for file in check_list(
                     task.get(
@@ -571,6 +585,8 @@ def add_tasks_to_task_group(
                     )
                 )
             ]
+
+            # Add TaskOutput to 'outputs'?
             if check_bool(
                 task.get(
                     CAPTURE_TASKOUTPUT,
@@ -582,10 +598,11 @@ def add_tasks_to_task_group(
                     ),
                 )
             ):
-                output_files.append(TaskOutput.from_task_process())
+                outputs.append(TaskOutput.from_task_process())
 
-            # If there's no task type in the task definition, and
-            # there's only one task type at the task group level, use it
+            # If there's no task type in the task definition, AND
+            # there's only one task type at the task group level,
+            # use that task type
             try:
                 task_type = task[TASK_TYPE]
             except KeyError:
@@ -604,9 +621,8 @@ def add_tasks_to_task_group(
                     executable=executable,
                     args=arguments_list,
                     env=env,
-                    inputs=input_files,
-                    outputs=output_files,
-                    uploaded_files=uploaded_files,
+                    inputs=inputs,
+                    outputs=outputs,
                     flatten_upload_paths=flatten_upload_paths,
                 )
             )
@@ -630,6 +646,9 @@ def add_tasks_to_task_group(
                     f"Task(s) to Work Requirement Task Group '{task_group.name}'"
                 )
 
+        # Empty the task list for the next batch
+        tasks_list.clear()
+
     if not ARGS_PARSER.dry_run:
         print_log(
             f"Added a total of {num_tasks} Task(s) to Task Group '{task_group.name}'"
@@ -673,27 +692,11 @@ def cleanup_on_failure(work_requirement: WorkRequirement) -> None:
     if ARGS_PARSER.dry_run:
         return
 
-    def _delete_objects():
-        object_paths: List[
-            ObjectPath
-        ] = CLIENT.object_store_client.get_namespace_object_paths(
-            ObjectPathsRequest(CONFIG_COMMON.namespace)
-        )
-        object_paths_to_delete: List[ObjectPath] = []
-        for object_path in object_paths:
-            if work_requirement.name in object_path.name:
-                object_paths_to_delete.append(object_path)
-        if len(object_paths_to_delete) > 0:
-            CLIENT.object_store_client.delete_objects(
-                CONFIG_COMMON.namespace, object_paths=object_paths_to_delete
-            )
-            print_log(f"Deleted all Objects under '{work_requirement.name}'")
-        else:
-            print_log("No Objects to Delete")
-
     CLIENT.work_client.cancel_work_requirement(work_requirement)
     print_log(f"Cancelled Work Requirement '{work_requirement.name}'")
-    _delete_objects()
+
+    # Delete uploaded objects
+    UPLOADED_FILES.delete_uploaded_files()
 
 
 def check_for_duplicates_in_file_lists(
@@ -811,7 +814,6 @@ def create_task(
     env: Dict[str, str],
     inputs: Optional[List[TaskInput]],
     outputs: Optional[List[TaskOutput]],
-    uploaded_files: List[str],
     flatten_upload_paths: bool = False,
 ) -> Task:
     """
@@ -836,21 +838,14 @@ def create_task(
         flatten_input_paths = FlattenPath.FILE_NAME_ONLY
 
     # Special processing for Bash tasks. The Bash script is uploaded if this
-    # hasn't already been done, and added it's to the list of required files.
+    # hasn't already been done, and it's added to the list of required files.
     if task_type == "bash":
         if executable is None:
             raise Exception("No 'executable' specified for 'bash' Task Type")
-        if executable not in uploaded_files and not ARGS_PARSER.dry_run:
-            upload_file(
-                client=CLIENT,
-                filename=executable,
-                id=ID,
-                namespace=CONFIG_COMMON.namespace,
-                url=CONFIG_COMMON.url,
-                input_folder_name=INPUT_FOLDER_NAME,
-                flatten_upload_paths=flatten_upload_paths,
-            )
-            uploaded_files.append(executable)
+        UPLOADED_FILES.add_input_file(
+            filename=executable,
+            flatten_upload_paths=flatten_upload_paths,
+        )
         task_input = TaskInput.from_task_namespace(
             unique_upload_pathname(
                 filename=executable,
@@ -950,13 +945,18 @@ def submit_json_raw(wr_file: str):
     input files. These can be pre-uploaded using yd-upload.
     """
 
-    # Load file contents, with Jsonnet/Mustache processing if required
+    # Load file contents, with Mustache processing
     if wr_file.lower().endswith(".jsonnet"):
         wr_data = load_jsonnet_file_with_mustache_substitutions(wr_file)
-    else:
+    elif wr_file.lower().endswith(".json"):
         wr_data = load_json_file_with_mustache_substitutions(wr_file)
+    else:
+        raise Exception(
+            f"Work Requirement file '{wr_file}' must end in '.json' or '.jsonnet'"
+        )
 
     if ARGS_PARSER.dry_run:
+        # This will show the results any Mustache processing
         print_log("Dry-run: Printing JSON Work Requirement specification:")
         print_json(wr_data)
         print_log("Dry-run: Complete")
@@ -990,37 +990,42 @@ def submit_json_raw(wr_file: str):
         raise Exception(f"{response.text}")
 
     # Submit Tasks to the Work Requirement
-    input_files = set()
+    verify_at_start_files = set()
     for task_group_name, task_list in task_lists.items():
 
-        # Collect set of required files
+        # Collect set of VERIFY_AT_START files
         for task in task_list:
-            input_files.update(
-                {input["objectNamePattern"] for input in task.get("inputs", [])}
-            )
+            for input in task.get("inputs", []):
+                if input["verification"] == "VERIFY_AT_START":
+                    namespace = (
+                        CONFIG_COMMON.namespace
+                        if input["source"] == "TASK_NAMESPACE"
+                        else input["namespace"]
+                    )
+                    verify_at_start_files.add(
+                        f"{namespace} :: {input['objectNamePattern']}"
+                    )
 
-        # Warn about required files & pause to allow upload
-        if len(input_files) != 0:
+        # Warn about VERIFY_AT_START files & halt to allow upload or
+        # Work Requirement cancellation
+        if len(verify_at_start_files) != 0:
             print_log(
                 "The following files may be required ('VERIFY_AT_START') "
                 "before Tasks are submitted, or the Tasks will fail."
             )
             print_log(
-                "You now have an opportunity to upload required files "
-                "before proceeding:\n"
+                "You now have an opportunity to upload the required files "
+                "before Tasks are submitted:\n"
             )
-            print_numbered_strings(sorted(list(input_files)))
-            if not confirmed(
-                "Now proceed with Task submission (Y/y), or "
-                "Cancel Work Requirement (N)?"
-            ):
+            print_numbered_strings(sorted(list(verify_at_start_files)))
+            if not confirmed("Proceed now (y), or Cancel Work Requirement (n)?"):
                 print_log(f"Cancelling Work Requirement '{wr_name}'")
                 CLIENT.work_client.cancel_work_requirement_by_id(wr_id)
                 return
 
         # Submit Tasks in batches
-        batches = ceil(len(task_list) / TASK_BATCH_SIZE)
-        for index in range(batches):
+        num_batches = ceil(len(task_list) / TASK_BATCH_SIZE)
+        for index in range(num_batches):
             task_batch = task_list[
                 index
                 * TASK_BATCH_SIZE : min(len(task_list), (index + 1) * TASK_BATCH_SIZE)
@@ -1040,7 +1045,7 @@ def submit_json_raw(wr_file: str):
             if response.status_code == 200:
                 print_log(
                     f"Added {len(task_batch)} Task(s) to Task Group "
-                    f"'{task_group_name}' (Batch {index + 1} of {batches})"
+                    f"'{task_group_name}' (Batch {index + 1} of {num_batches})"
                 )
             else:
                 print_error(f"Failed to add Task(s) to Task Group '{task_group_name}'")
@@ -1052,36 +1057,6 @@ def submit_json_raw(wr_file: str):
         follow_progress(CLIENT.work_client.get_work_requirement_by_id(wr_id))
 
 
-def csv_expand_toml_tasks(csv_file: str) -> Dict:
-    """
-    When there's a CSV file specified, but no JSON file, create the expanded
-    list of Tasks using the CSV data.
-    """
-    wr_data = {TASK_GROUPS: [{TASKS: [{}]}]}
-    task_proto = wr_data[TASK_GROUPS][0][TASKS][0]
-    for config_value, config_name in [
-        (CONFIG_WR.args, ARGS),
-        (CONFIG_WR.bash_script, BASH_SCRIPT),
-        (CONFIG_WR.capture_taskoutput, CAPTURE_TASKOUTPUT),
-        (CONFIG_WR.docker_env, DOCKER_ENV),
-        (CONFIG_WR.docker_password, DOCKER_PASSWORD),
-        (CONFIG_WR.docker_username, DOCKER_USERNAME),
-        (CONFIG_WR.env, ENV),
-        (CONFIG_WR.executable, EXECUTABLE),
-        (CONFIG_WR.flatten_input_paths, FLATTEN_PATHS),
-        (CONFIG_WR.input_files, INPUT_FILES),
-        (CONFIG_WR.wr_name, NAME),
-        (CONFIG_WR.output_files, OUTPUT_FILES),
-        (CONFIG_WR.task_type, TASK_TYPE),
-        (CONFIG_WR.verify_at_start, VERIFY_AT_START),
-        (CONFIG_WR.verify_wait, VERIFY_WAIT),
-    ]:
-        if config_value is not None:
-            task_proto[config_name] = config_value
-
-    return perform_csv_task_expansion(wr_data, [csv_file])
-
-
 # Standalone entry point
 if __name__ == "__main__":
     main()
