@@ -5,7 +5,7 @@ Report version numbers, etc.
 """
 
 from os.path import abspath
-from sys import executable, path
+from sys import argv, executable, path
 from sys import version as py_version
 
 from yellowdog_client._version import __version__ as yd_sdk_version
@@ -29,10 +29,11 @@ def main():
     print(f"  Jsonnet Support:                     {check_for_jsonnet()}")
     print(f"  YellowDog SDK Version:               {yd_sdk_version}")
     print(f"  Python Version:                      {py_version.split()[0]} ")
-    print(f"  Command:                             {abspath(__file__)}")
-    print(f"  Python Executable:                   {executable}")
-    for i, p in enumerate(path, start=1):
-        print(f"    Path-{str(i).zfill(2)}:                             {p}")
+    if "--debug" in argv:
+        print(f"  Command:                             {abspath(__file__)}")
+        print(f"  Python Executable:                   {executable}")
+        for i, p in enumerate(path, start=1):
+            print(f"    Path-{str(i).zfill(2)}:                           {p}")
 
 
 if __name__ == "__main__":
